/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// Module Name: rtavr_s1_decode
//   Ver 005
//
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                NORMAL
// Number of Slice Flip Flops:       24
// Number of 4 input LUTs:          161     
// Number of occupied Slices:        90    
// Total Number of 4 input LUTs:    173   
// Number of bonded IOBs :          144  
//   IOB Flip Flops :                42 
// Number of BUFGMUXs :               1
// Number of RAMB16BWEs :             - 
// Clock to Setup 
//on destination clock CLK         4.554

module rtavr_s1_decode (
		input CLK,
		input RESET,
		input [15:0] INST,
		input [11:0] JA,

		input S0_WB_INST,
`ifdef NO_STALL_IDX
`else
		input S0_IDX_INST,
`endif
		output S0_VALID,

		output CMD_LOAD, CMD_STORE,
		output CMD_LOAD_WR,
		output CMD_OP3, CMD_OP2, CMD_OP1,
		output [2:0] OP3_SEL,
		output OP3_WC, CMD_OP3_WR, CMD_OP3_WRF, 
		output CMD_OP2_WR, 
		output [2:0] OP2_SEL,
		output [2:0] OP1_SEL,
		output [7:0] IMM_DATA,
		input  [7:0] LOAD_DATA, // POP data

		output CMD_SBIX,
		output CMD_SBRX,
		output CMD_XBI,
		output CMD_BST,
		output CMD_BLD,
		output [2:0] SBIX_BIT,
		input  SBIX_BIT_IN,
		input  FLAGS_BIT_IN,
		output XBI_BIT_OUT,

		output EA_PUSH,
		output PUSH_SEL,
		output CMD_SETI,
		output CMD_CLRI,

		input [4:0] INT_VEC,
		input INT_REQ,
		output INT_ACK,

		output [15:0] SP_OUT,
		input  [15:0] SP_IN,
		input  [15:0] EA_INDEX,
		output [15:0] EA,

		output [11:0] PC
		);
  reg n_reset;
  reg s1_invalid;
  reg i1_int;
  reg i2_int;
  reg i2_call;
  reg i2_ret;
  reg i3_ret;
  reg i2_reti;
  reg i3_reti;

  reg s2_cmd_load;
  reg s2_cmd_load_wr;
  reg s2_cmd_store;
  reg s2_cmd_op3;
  reg s2_cmd_op3_wr;
  reg s2_cmd_op3_wrf;
  reg s2_cmd_op2;
  reg s2_cmd_op2_wr;
  reg s2_cmd_op1;
  reg s2_cmd_bst;
  reg s2_cmd_bld;
  reg s2_cmd_xbi;
  reg [2:0] s2_op3_sel;
  reg s2_op3_wc;
  reg [2:0] s2_op2_sel;
  reg [2:0] s2_op1_sel;
  reg s2_xbi_bit_out;

  reg s2_ea_push;
  reg s2_ea_pop;
  reg s2_push_sel;

  reg [15:0] s2_ea;

  reg s2_cpse;
  reg s2_sbix;
  reg s2_sbix_sel;
  reg [2:0] s2_sbix_bit;
  reg s2_sbrx;
`ifdef SUPPORT_LDST32
  reg s2_ldst32;
`endif

  reg [7:0] r_imm;	// op2 imm or push data
  reg [7:0] r_tmp;	// prev pcl or
  reg [11:0] r_pc;
 
  assign CMD_LOAD = s2_cmd_load;
  assign CMD_LOAD_WR = s2_cmd_load_wr;
  assign CMD_STORE = s2_cmd_store;

  assign CMD_OP3 = s2_cmd_op3;
  assign CMD_OP3_WR = s2_cmd_op3_wr;
  assign CMD_OP3_WRF = s2_cmd_op3_wrf;
  assign OP3_SEL = s2_op3_sel;
  assign OP3_WC = s2_op3_wc;
  
  assign CMD_OP2 = s2_cmd_op2;
  assign CMD_OP2_WR = s2_cmd_op2_wr;
  assign OP2_SEL = s2_op2_sel;
  assign CMD_OP1 = s2_cmd_op1;
  assign OP1_SEL = s2_op1_sel;

  assign EA_PUSH = s2_ea_push;
  assign PUSH_SEL = s2_push_sel;

  assign CMD_SBIX = s2_sbix;
  assign CMD_SBRX = s2_sbrx;
  assign SBIX_BIT = s2_sbix_bit;
  
  assign IMM_DATA = r_imm;
  assign PC = r_pc;

  wire f_ext = (INST[15:9] == 7'b1001010 );

// EA_ABS / ABS_ADDR
// IN              1011:0AA1:dddd:AAAA     # 00
// OUT             1011:1AA1:rrrr:AAAA     # 00
// LDS             1010:0kkK:dddd:kkkk     # 
// STS             1010:1kkK:dddd:kkkk     # 
//                 101X:S--X----------     # 
//  INST[12]/INST[8]  -> v_ea_abs[7:6]
//                  01        10
//                  00        01
//                  10        11 (undefined)
//                  11        00
// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #
// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     # 
//
// BSET            1001:0100:0sss:1000     # 
// BCLR            1001:0100:1sss:1000     # 
  

`ifdef S1_NO_BSET
  wire f_bset_bclr = 0;
`else
  wire f_bset_bclr = f_ext & ~INST[8] & (INST[3:0] == 4'b1000) ;
`endif

  wire v_ea_abs = (INST[15:13] == 3'b101)
		  | (INST[15:10] == 6'b100110)
		  | f_bset_bclr ;

  wire [7:0] v_ea = (INST[15:13] == 3'b101)
		? { INST[12] ^ INST[8] ,  ~INST[8] , INST[10:9] , INST[3:0] }
		: f_bset_bclr ? 8'h3f
		: { 3'b000 , INST[7:3] } ;

 wire f_sbix = (INST[15:10] == 6'b100110) & INST[8];
 
// # OP3 INST
// NOP		   0000:0000:0000:0000	   #
// CPSE            000B:00rd:dddd:rrrr     # not LOAD 
// CP/CPC          000B:01rd:dddd:rrrr     #
// SUB/SBC         000B:10rd:dddd:rrrr     #
// ADD/ADC         000C:11rd:dddd:rrrr     # 
// AND             0010:00rd:dddd:rrrr     # 
// EOR             0010:01rd:dddd:rrrr     # 
// OR              0010:10rd:dddd:rrrr     # 
// MOV             0010:11rd:dddd:rrrr     # 
//                 --*-:**------------
// CPI             0011:KKKK:dddd:KKKK     # 
// MOVW            0000:0001:dddd:rrrr     # undefined inst
// MULS            0000:0010:dddd:rrrr     # undefined inst
// MULSU           0000:0011:0ddd:0rrr     # undefined inst
// FMUL            0000:0011:0ddd:1rrr     # undefined inst
// FMULS           0000:0011:1ddd:0rrr     # undefined inst
// FMULSU          0000:0011:1ddd:1rrr     # undefined inst

  wire v_cmd_op3 = (INST[15:14] == 2'b00) & ~(INST[13:12] == 2'b11);
  wire v_cmd_op3_wr = v_cmd_op3 & (INST[13] | INST[11]);
  wire v_cmd_op3_wrf = v_cmd_op3 & (INST[13] | ~(INST[11:10] == 2'b00));
  wire f_cpse = v_cmd_op3 & ~INST[13] & (INST[11:10] == 2'b00);
  wire [2:0] v_op3_sel = { INST[13] , INST[11:10] };
  wire v_op3_wc = INST[12];

  wire f_skip = ( s2_cpse | s2_sbrx | s2_sbix ) & s2_sbix_sel & SBIX_BIT_IN
               | ( s2_sbrx | s2_sbix ) & ~s2_sbix_sel & ~SBIX_BIT_IN ;
 
// # OP2 INST
// LDI             1110:KKKK:dddd:KKKK     # OP2_SEL = 2
// CPI             0011:KKKK:dddd:KKKK     # OP2_SEL = 3
// SBCI            0100:KKKK:dddd:KKKK     # 
// SUBI            0101:KKKK:dddd:KKKK     # 
// ORI             0110:KKKK:dddd:KKKK     # 
// ANDI            0111:KKKK:dddd:KKKK     # 
//                 -***:--------------      

  wire v_cmd_op2 = (INST[15:14] == 2'b01 )
			| (INST[15:12] == 4'b0011)
			| (INST[15:12] == 4'b1110) ;
  wire v_cmd_op2_wr = v_cmd_op2 & ~(INST[14:12] == 3'b011);

  wire [2:0] v_op2_sel = { (INST[15] ^ INST[14]), INST[13:12] };
  wire [7:0] v_imm = { INST[11:8] , INST[3:0] };

// # OP1 INST
// COM             1001:010d:dddd:0000     # 
// NEG             1001:010d:dddd:0001     # 
// SWAP            1001:010d:dddd:0010     # 
// INC             1001:010d:dddd:0011     # 
// DEC             1001:010d:dddd:1010     # OP1_SEL = 4
// ASR             1001:010d:dddd:0101     # 
// LSR             1001:010d:dddd:0110     # 
// ROR             1001:010d:dddd:0111     # 
//                 ----:-----------***      

  wire v_cmd_op1 = f_ext & (~INST[3] | (INST[3:0] == 4'b1010));

  wire [2:0] v_op1_sel = (INST[3]) ? 3'b100 : INST[2:0];

// LOAD Rr (Rd -> Rr @s0)
// LD.Rd.Z         1000:000d:dddd:0000     # 
// LD.Rd.Y         1000:000d:dddd:1000     # 
// LD.Rd.Z++       1001:000d:dddd:0001     # 
// LD.Rd.--Z       1001:000d:dddd:0010     # 
// LD.Rd.Y++       1001:000d:dddd:1001     # 
// LD.Rd.--Y       1001:000d:dddd:1010     # 
// LD.Rd.X         1001:000d:dddd:1100     # 
// LD.Rd.X++       1001:000d:dddd:1101     # 
// LD.Rd.--X       1001:000d:dddd:1110     # 

// POP             1001:000d:dddd:1111     # 

// SYORE Rr
// ST.Z.Rr         1000:001r:rrrr:0000     # 
// ST.Y.Rr         1000:001r:rrrr:1000     # 
// ST.Z++.Rr       1001:001r:rrrr:0001     # 
// ST.--Z.Rr       1001:001r:rrrr:0010     # 
// ST.Y++.Rr       1001:001r:rrrr:1001     # 
// ST.--Y.Rr       1001:001r:rrrr:1010     # 
// ST.X.Rr         1001:001r:rrrr:1100     # 
// ST.X++.Rr       1001:001r:rrrr:1101     # 
// ST.--X.Rr       1001:001r:rrrr:1110     # 

// PUSH            1001:001d:dddd:1111     # 
//                 100X:00S-----------

// LPM.Rd.Z        1001:000d:dddd:0100     # undefined inst
// LPM.Rd.Z++      1001:000d:dddd:0101     # undefined inst
// ELPM.Rd.Z       1001:000d:dddd:0110     # undefined inst
// ELPM.Rd.Z++     1001:000d:dddd:0111     # undefined inst
// LDS(2word)      1001:000d:dddd:0000     # 
// STS(2word)      1001:001d:dddd:0000     # 

  // LD/ST + POP/PUSH
  wire v_ldst = (INST[15:13] == 3'b100) & (INST[11:10] == 2'b00);
`ifdef SUPPORT_LDST32
  wire v_ldst32 = v_ldst & INST[12] & (INST[3:0] == 4'b0000);
`endif
  
  
// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #
// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     # 

  wire v_ldio = (INST[15:10] == 6'b100110);
//  wire v_stio = v_ldio & ~INST[8];

// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #

// BSET            1001:0100:0sss:1000     # 
// BCLR            1001:0100:1sss:1000     # 

`ifdef S1_NO_XBI
  wire v_cmd_xbi = 0;
`else
  wire v_cmd_xbi = v_ldio & ~INST[8] | f_bset_bclr;
`endif

  wire v_xbi_bit_out = f_bset_bclr ? ~INST[7] : INST[9];
  assign XBI_BIT_OUT = s2_xbi_bit_out;
  assign CMD_XBI = s2_cmd_xbi;

//
// IN              1011:0AA1:dddd:AAAA     # 
// LDS             1010:0kkK:dddd:kkkk     # 
// OUT             1011:1AA1:rrrr:AAAA     # 
// STS             1010:1kkK:dddd:kkkk     #
//                 101-:S-------------

  wire v_cmd_load_wr = ( v_ldst & ~INST[9] )    // LD,POP
                    | ( v_ea_abs & ~INST[11] ); // IN, LDS

  wire v_cmd_load = ( v_ldst & ~INST[9] )       // LD,POP
                    | ( v_ea_abs & ~INST[11] )  // IN, LDS
  		    |  v_ldio ;                 // CBI,SBI , SBIC,SBIS
		    
  wire v_cmd_store = ( v_ldst & INST[9] )       // ST,PUSH
                    | ( v_ea_abs & INST[11] )   // OUT, STS
		    |  v_cmd_xbi ;              // CBI,SBI


// IJMP            1001:0100:0000:1001     # 
// ICALL           1001:0101:0000:1001     # 
//                 EEEE:EEE*:----:1--1     # 

  wire f_ij =  f_ext & (~INST[8] & INST[3] & INST[0]);
  wire f_ic =  f_ext & (INST[8] & INST[3] & INST[0]);

// RJMP            1100:kkkk:kkkk:kkkk     # 
// RCALL           1101:kkkk:kkkk:kkkk     # 

  wire [11:0] adr_rj = INST[11:0];
  wire f_rj =  (INST[15:12] == 4'b1100 );
  wire f_rc =  (INST[15:12] == 4'b1101 );

// PUSH            1001:001d:dddd:1111     # 
// ICALL           1001:0101:0000:1001     # 
// RCALL           1101:kkkk:kkkk:kkkk     # 

// POP             1001:000d:dddd:1111     # 
// RET             1001:0101:0000:1000    # 
// RETI            1001:0101:0001:1000    # 
//                 EEEE:EEE1:0--*:1--0    #

  wire f_ret = f_ext & INST[8]
			& (INST[3:0] == 4'b1000);	// RET/RETI
  wire f_reti = f_ret & INST[4];                        // is RETI
  assign CMD_SETI = i3_reti;

  wire v_push_pop = v_ldst & (INST[3:0] == 4'b1111);
  wire v_ea_pop = ( v_push_pop & ~INST[9] )   // POP;
                   | f_ret ;                  // RET,RETI

  wire v_ea_push = ( v_push_pop & INST[9] )   // PUSH;
                   | f_rc | f_ic              // RCALL,ICALL
                   | i1_int;		      // interrupt

  wire v_push_sel =  f_rc | f_ic | i2_call | f_ret | i2_ret  // 1: PC 0: GPR
		   | i1_int;

  wire [11:0] v_ret_addr = { LOAD_DATA[3:0] , r_tmp };

// BLD             1111:100d:dddd:0bbb     #
// BST             1111:101d:dddd:0bbb     # 

  wire v_cmd_bst = (INST[15:9] == 7'b1111101 );
`ifdef S1_NO_BLD
  wire v_cmd_bld = 0;
`else
  wire v_cmd_bld = (INST[15:9] == 7'b1111100 );
`endif
  assign CMD_BLD = s2_cmd_bld;
  assign CMD_BST = s2_cmd_bst;

// BRBC            1111:01kk:kkkk:ksss     # 
// BRBS            1111:00kk:kkkk:ksss     #

  wire [11:0] adr_br = { INST[9], INST[9], INST[9], INST[9], INST[9]
			 ,INST[9:3] };
  wire f_br = (INST[15:11] == 5'b11110 ) & ( INST[10] ^ FLAGS_BIT_IN);

// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     #

// SBRC            1111:110r:rrrr:0bbb     #
// SBRS            1111:111r:rrrr:0bbb     #

// BSET            1001:0100:0sss:1000     # 
// BCLR            1001:0100:1sss:1000     # 

  wire f_sbrx = (INST[15:10] == 6'b111111 );
  wire f_sbix_sel = INST[9]; // 0 : SBIC/SBRC 1: SBIS/SBRS
  wire [2:0] f_sbix_bit = f_bset_bclr ? INST[6:4] : INST[2:0];
  
// SLEEP           1001:0101:1000:1000     #
// WDR             1001:0101:1010:1000     #
// BREAK           1001:0101:1001:1000     # undefined inst
//                 EEEE:EEE1:1-**:1--0     #

  wire f_mcu = f_ext & (INST[8:7] == 2'b11) & INST[3] & ~INST[0];
  wire f_sleep = f_mcu & (INST[5:4] == 2'b00);
  wire f_wdr   = f_mcu & (INST[5:4] == 2'b10);


  // SP_OUT is s2 op
  wire [11:0] sp_dec = SP_IN[11:0] - 12'h001;
  wire [11:0] sp_inc = SP_IN[11:0] + 12'h001;
  assign SP_OUT[15:12] = SP_IN[15:12];
  assign SP_OUT[11:0] = ( s2_ea_push ) ? sp_dec
                        : ( s2_ea_pop ) ? sp_inc
		        : SP_IN[11:0];

  // EA is s1 op
  wire [15:0] v2_ea = (v_ea_push | i2_call) ? { 4'h0, SP_IN }
                   : (v_ea_pop | i2_ret)  ? { 4'h0, sp_inc }
		   : v_ea_abs  ? { 8'h00, v_ea }
		   : EA_INDEX;
`ifdef SUPPORT_LDST32
  assign EA = s2_ldst32 ? INST : s2_ea;
`else
  assign EA = s2_ea;
`endif

  wire [11:0] v_pc = 
		     i1_int ? { 7'b0000000, INT_VEC }
		   : ~s1_invalid & (f_ij | f_ic)  ? JA 
                   : i3_ret ? v_ret_addr
                   : r_pc + ( ~s1_invalid & (f_rj | f_rc) ? adr_rj 
                                :  ~s1_invalid & f_br ? adr_br 
		                : 1);

`ifdef NO_STALL_IDX
  wire s0_inv_conflict = (S0_WB_INST)  & ~s1_invalid
                  & (v_cmd_op3_wr | v_cmd_op2_wr | v_cmd_op1 
		    | v_cmd_bld | v_cmd_load_wr);
`else
  wire s0_inv_conflict = (S0_WB_INST | S0_IDX_INST)  & ~s1_invalid
                  & (v_cmd_op3_wr | v_cmd_op2_wr | v_cmd_op1 
		    | v_cmd_bld | v_cmd_load_wr);
`endif

  wire s0_inv_jump = ~s1_invalid & ( f_ij | f_ic | f_br | f_rj | f_rc )
		      | i1_int | i3_ret;

`ifdef SUPPORT_LDST32
  wire s0_inv_skip = ~s1_invalid & (f_skip | f_ret | v_ldst32) | i2_ret;
`else
  wire s0_inv_skip = ~s1_invalid & (f_skip | f_ret) | i2_ret;
`endif

  wire v_int  = ~(s0_inv_skip | s0_inv_jump) & INT_REQ & ~CMD_SETI ;

  assign S0_VALID = ~( s0_inv_conflict | s0_inv_jump | s0_inv_skip
			| v_int);
                  
  wire v_load_pc = ~(n_reset | s0_inv_conflict | v_int );
//  wire v_load_pc = ~(s0_inv_conflict | v_int );

  wire v_invalid =  (n_reset | ~S0_VALID);
//  wire v_invalid =  (~S0_VALID);

  assign CMD_CLRI = i1_int;
  assign INT_ACK = i2_int;

  always @(negedge CLK)
  begin
    if (RESET == 0) 
      begin
	n_reset       <= 1;
	s1_invalid    <= 1;
	i1_int         <= 0;
	i2_int         <= 0;

	r_pc          <= 0;
	r_imm         <= 0;
	r_tmp         <= 0;

	s2_ea_push     <= 0;
	s2_ea_pop      <= 0;
	s2_push_sel    <= 0;
	s2_ea         <= 0;

	s2_cpse        <= 0;
	s2_sbix        <= 0;
	s2_sbrx        <= 0;
	s2_sbix_sel    <= 0;
	s2_sbix_bit    <= 0;

	s2_cmd_op3     <= 0;
	s2_cmd_op3_wr  <= 0;
	s2_cmd_op3_wrf <= 0;
	s2_op3_sel     <= 0;
	s2_op3_wc      <= 0;

	s2_cmd_op2     <= 0;
	s2_cmd_op2_wr  <= 0;
	s2_op2_sel     <= 0;

	s2_cmd_op1     <= 0;
	s2_op1_sel     <= 0;

	s2_cmd_load    <= 0;
	s2_cmd_load_wr <= 0;
	s2_cmd_store   <= 0;
	s2_cmd_bld     <= 0;
	s2_cmd_bst     <= 0;
	s2_cmd_xbi     <= 0;
	s2_xbi_bit_out <= 0;
`ifdef SUPPORT_LDST32
	s2_ldst32      <= 0;
`endif

	i2_call        <= 0;
	i2_ret         <= 0;
	i3_ret         <= 0;
	i2_reti        <= 0;
	i3_reti        <= 0;
      end
    else
      begin
	n_reset       <=  0;
	s1_invalid    <=  v_invalid | v_int;
	i1_int        <= v_int;
	i2_int        <= i1_int;
	if (v_load_pc & ~v_int) r_pc <= v_pc;

	r_imm         <= v_ea_push ? {4'h0, r_pc[11:8]} 
			: i2_call ? r_tmp : v_imm;
	r_tmp         <= i2_ret ? LOAD_DATA : r_pc[7:0];

	s2_ea_push     <= v_ea_push | i2_call;
	s2_ea_pop      <= v_ea_pop | i2_ret;
	s2_push_sel    <= v_push_sel;
	s2_ea          <= v2_ea;

	s2_cpse        <= f_cpse;
	s2_sbix        <= f_sbix;
	s2_sbrx        <= f_sbrx;
	s2_sbix_sel    <= f_sbix_sel;
	s2_sbix_bit    <= f_sbix_bit;

	s2_cmd_op3     <= ~s1_invalid & v_cmd_op3;
	s2_cmd_op3_wr  <= ~s1_invalid & v_cmd_op3_wr;
	s2_cmd_op3_wrf <= ~s1_invalid & v_cmd_op3_wrf;
	s2_op3_sel     <= v_op3_sel;
	s2_op3_wc      <= v_op3_wc;

	s2_cmd_op2     <= ~s1_invalid & v_cmd_op2;
	s2_cmd_op2_wr  <= ~s1_invalid & v_cmd_op2_wr;
	s2_op2_sel     <= v_op2_sel;

	s2_cmd_op1     <= ~s1_invalid & v_cmd_op1;
	s2_op1_sel     <= v_op1_sel;

	s2_cmd_load    <= (~s1_invalid & (v_cmd_load | f_ret)) | i2_ret;
	s2_cmd_load_wr <= ~s1_invalid & v_cmd_load_wr ;
	s2_cmd_store   <= (~s1_invalid & (v_cmd_store | f_ic | f_rc))
				 | i2_call | i1_int;

	s2_cmd_bld     <= ~s1_invalid & v_cmd_bld;
	s2_cmd_bst     <= ~s1_invalid & v_cmd_bst;
	s2_cmd_xbi     <= ~s1_invalid & v_cmd_xbi;
	s2_xbi_bit_out <= v_xbi_bit_out;
`ifdef SUPPORT_LDST32
	s2_ldst32      <=  ~s1_invalid & v_ldst32;
`endif

	i2_call        <=  ~s1_invalid & (f_rc | f_ic)  // RCALL,ICALL (2)
				| i1_int;		// interrupt

	i2_ret         <=  ~s1_invalid & f_ret;         // RET/RETI    (2)
	i3_ret         <=  i2_ret;                      // RET/RETI    (3)
	i2_reti        <=  ~s1_invalid & f_reti;        // RETI        (2)
	i3_reti        <=  i2_reti;                     // RETI        (3)
      end
  end

endmodule
